#                                                         -*- Perl -*-
# Wdic.pm - WDIC manager, to be used from wdic-fpw.
# $Id: Wdic.pm 1.2.0.8 2001/01/25 00:32:06 satomii Exp $
#
# KANJI-CODE: SJIS; TAB-SIZE: 4
# DO NOT CHANGE THE KANJI CODE OF THIS SCRIPT. IT MUST BE SHIFT JIS.
#
# Copyright (C) 2000, rei@wdic.org.
# This program is distributed in the hope that it will be useful, but
# without any warranty. See the GNU General Public License for the details.
#
package Wdic;

require 5.005;
require Exporter;

use IO::File;
use File::DosGlob 'glob';
use integer;
use strict;

use vars qw(
	$VERSION @ISA @EXPORT @EXPORT_OK
);

$VERSION = '1.2';

@ISA = qw(Exporter);
@EXPORT = qw(
	&Wdic::init_wdic($)
	&Wdic::uninit_wdic()
	&Wdic::get_group_list()
	&Wdic::get_group_info($)
	&Wdic::get_file_title($)
	&Wdic::locate_word($$$)
	&Wdic::define_word($)
);
@EXPORT_OK = qw(
	$TRACE
	$CHARSET
);

my $TRACE = 1;		# 1 for status messages, 2 for debug messages.
my $CHARSET = '';	# display charset for tracing.

use fields qw(
	directory
	wdic_group
	wdic_file
	word_list
	word_index
);

my $SJIS_ANY = '(?:[\x81-\x9f\xe0-\xfc][\x40-\x7e\x80-\xfc])';
my $SJIS_YOMI =
	'(?:(?:\x82[\x4f-\x58\x60-\x79\x82-\x9a])' .	# alnum.
	'|(?:\x82[\x9f-\xf1])' .						# hiragana.
	'|(?:\x83[\x40-\x7e\x80-\x96]))';				# katakana.

#-----------------------------------------------------------------------#
#	constructor and destructor											#
#-----------------------------------------------------------------------#

sub new()
{
	my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = {
		directory => '',
		wdic_group => {},
		wdic_file => {},
		word_list => []
	};
	bless($self, $class);
	$self->initialize($_[0]) if (@_);
	return($self);
}

sub DESTROY()
{
	# leave everything to GC...
}

#-----------------------------------------------------------------------#
#	exported methods													#
#-----------------------------------------------------------------------#
#
# int init_wdic(char *directory)
# loads WDIC files from the specified directory and constructs the
# word list to be used later by other methods. returns the number of
# words indexed.
#
sub Wdic::init_wdic($)
{
	my $self = shift;
	my $dir = $_[0];
	my(%wdic_group, %wdic_file);
	my(@list, $path, %entry);

	$dir =~ s/^\\/\//g;
	$dir =~ s/\/$//;
	-d $dir || die "'$dir' is not a valid directory ($!)";

	$self->uninit_wdic() if ('' ne $self->{directory});
	$self->{directory} = $dir;

	# load the group files if any.
	@list = glob("$dir/*.WLF");
	warn "The file '$dir/*.WLF' could not be found" if (!@list);
	foreach $path (@list) {
		eval { $self->load_group($path) };
		warn $@ if ($@);
	}

	# load wdic files and construct word list.
	@list = glob("$dir/*.DIC");
	die "The file '$dir/*.DIC' could not be found" if (!@list);
	foreach $path (@list) {
		eval { $self->load_file($path) };
		warn $@ if ($@);
	}

	@list = @{ $self->{word_list} };
	die "No word is found" if (!scalar(@list));

	# sort word entries.
	&trace("Sorting " . scalar(@list) . " entries...");
	@list = sort {
		my %a = %{ $a };
		my %b = %{ $b };
		$a{SORT} cmp $b{SORT} || $a{WORD} cmp $b{WORD};
	} @list;
	$self->{word_list} = \@list;
	&trace(" done.\n");

	# make simple index for the word list.
	&trace("Making index...");
	$self->make_word_index();
	&trace(" done.\n");

	if (2 <= $TRACE) {
		&trace("DEBUG: Wdic::initialize() is dumping class fields...\n");
		&trace('$directory = ' . $self->{directory} . "\n");
		&dump_hash(0, 'wdic_group', %{ $self->{wdic_group} });
		&dump_hash(0, 'wdic_file', %{ $self->{wdic_file} });
	}
	return scalar(@{ $self->{word_list} });
}

#
# void uninit_wdic(void)
#
sub Wdic::uninit_wdic()
{
	my $self = shift;

	$self->{directory} = '';
	undef $self->{wdic_group};
	undef $self->{wdic_file};
	undef $self->{word_list};
}

#
# char **get_group_list(void)
# returns the group names currnetly loaded into this module.
#
sub Wdic::get_group_list()
{
	my $self = shift;
	my @list = keys(%{ $self->{wdic_group} });
	my($file, %order, $handle);

	# try to open the group list file, FILE.GL.
	$file = glob($self->{directory} . '/FILE.GL');
	$handle = new IO::File($file, 'r') if ($file);

	if (defined($handle)) {
		my($str, $count);

		binmode $handle;
		$count = 0;
		while ($str = $handle->getline()) {
			if ($str =~ /^\[GROUP\]/i) {
				last if ($count);
				$count++;
				next;
			} elsif (!$count) {
				next;
			}
			$str =~ s/\#.*$//;
			$str =~ s/[\r\n]+$//;
			if ($str =~ s/^.*\t([^\.]+)\.WLF\t.*$/$1/) {
				$order{$str} = $count++;
			}
		}
		undef $handle;

	} else {
		warn "Unable to open '" . $self->{directory} . "/FILE.GL' ($!)";
		%order = (WDIC => 1, TECH => 2, MOE => 3, RAIL => 4);
	}

	# sort the group list.
	@list = sort {
		if (defined($order{$a})) {
			defined($order{$b}) ? $order{$a} <=> $order{$b} : 1;
		} elsif (defined($order{$b})) {
			-1;
		} else {
			$a cmp $b;
		}
	} @list;

	if (2 <= $TRACE) {
		&dump_array(0, 'group_list', @list);
	}

	return @list;
}

#
# (hash) get_group_info(char *group_name)
# returns a hash which contains group information.
#
sub Wdic::get_group_info($)
{
	my $self = shift;
	my %group = %{ $self->{wdic_group} };

	return if ($_[0] !~ /^[A-Z]+$/i);
	%group = %{ $group{$_[0]} };
	$group{NAME} = $_[0] if ('' eq $group{NAME});
	if (2 <= $TRACE) {
		&dump_hash(0, $_[0], %group);
	}
	return %group;
}

#
# char *get_group_title(char *group)
# returns the title of a group. if the group was not found in the group
# list, the method returns the same string as passed.
#
sub Wdic::get_group_title($)
{
	my $self = shift;
	my $key = $_[0];
	my %group = %{ $self->{wdic_group} };

	if (%group && defined($group{$key})) {
		%group = %{ $group{$key} };
		return $group{NAME} if ($group{NAME});
	}
	return $key;
}

#
# char *get_file_title(char *file)
# returns the title of a file (dictionary). if the file was not found
# in the file list, the method returns the same string as passed.
#
sub Wdic::get_file_title($)
{
	my $self = shift;
	my $key = $_[0];
	my %file = %{ $self->{wdic_file} };

	if (%file && defined($file{$key})) {
		%file = %{ $file{$key} };
		return $file{NAME} if ($file{NAME});
	}
	return $key;
}

#
# char **locate_word(char *word, char *group, char *file)
# finds the specified word from the index and returns the group, file
# and offset in the array context.
#
sub Wdic::locate_word($$$)
{
	my $self = shift;
	my($word, $group, $file) = @_;
	my(%word_index, %entry, @word_list, @list, $key);

	if ($group) {
		# if the specified group is not available, immediately return.
		return if (!defined(${%{ $self->{wdic_group} }}{$group}));
	}

	%word_index = %{ $self->{word_index} };
	@list = unpack('SS', $word);
	$key = sprintf("%04X", $list[0]);
	$key .= sprintf("%04X", $list[1]) if (defined($list[1]));

	return if (!defined($word_index{$key}));
	@list = @{ $word_index{$key} };
	@word_list = @{ $self->{word_list} };

	foreach $key (@list) {
		%entry = %{ $word_list[$key] };
		next if ('' ne $file && $file ne $entry{FILE});
		next if ('' ne $group && $group ne $entry{GROUP});
		if ($word eq $entry{WORD}) {
			return ($key, $entry{GROUP}, $entry{FILE}, $entry{OFFSET});
		}
	}
	0;
}

#
# (hash) define_word(int index)
# returns one word definition at the specified index.
#
# %word = {
#     WORD => $word,
#     READ => $reading,
#     ENGLISH => $english,
#     PRONOUNCE => $pronounce,
#     POS => $part_of_speech,
#     DEFINITION => @definitions,
#     COLUMN => @columns_or_examples,
#     REFERENCE => @references,
#     GROUP => $group,
#     FILE => $file
# };
#
sub Wdic::define_word($)
{
	my $self = shift;
	my $index = $_[0];
	my(%word, %file, %define, @info, @def, @col, @ref, $str, $info);

	$index < scalar(@{ $self->{word_list} })
		|| die "Invalid word index $index";
	%word = %{ ${@{ $self->{word_list} }}[$index] };
	%file = %{ ${%{ $self->{wdic_file} }}{$word{FILE}} };

	# seek to the word position and validate the word there.
	$file{HANDLE}->seek($word{OFFSET}, IO::File::SEEK_SET)
		|| die "$word{FILE}.DIC: Failed to seek to offset $word{OFFSET} ($!)";
	$str = $file{HANDLE}->getline()
		|| die "$word{FILE}.DIC: File read error ($!)";
	if ($str !~ s/^\#([^\r\n]+)[\r\n]+$/$1/ || $str ne $word{WORD}) {
		die "$word{FILE}.DIC: $word{WORD}: Invalid offset $word{OFFSET}";
	}

	# read the next some lines.
	$info = '';
	while ($str = $file{HANDLE}->getline()) {
		last if ($str !~ s/^ {8}([^\r\n]+)[\r\n]+$/$1/);
		if ($str =~ /^E/) {
			push(@def, $str);
			last;
		} elsif ($str =~ /^/) {
			push(@ref, $str);
			last;
		} else {
			$info .= $str;
		}
	}
	@info = &split_info($info);

	# read body lines of this word.
	while ($str = $file{HANDLE}->getline()) {
		last if ($str !~ s/^ {8}([^\r\n]*)[\r\n]+$/$1/);
		if (@ref) {
			if ($str =~ s/^  //) {
				$ref[scalar(@ref) - 1] .= $str;
			} else {
				push(@ref, $str);
			}
		} elsif ($str =~ /^\x81[\xa8\x97\x9a\x9f\xa5\xa4]/) {
			# []
			push(@ref, $str);
		} elsif (@col) {
			push(@col, $str);
		} elsif ($str =~ /^E/) {
			# "E...", should be a description line.
			push(@def, $str);
		} elsif ($str =~ s/^  ([^ ])/$1/) {
			# a description line, to be combined to the previous line.
			$str =~ s/ +$//;
			$str =~ s/  +/ /;
			if ($def[$#def] =~ /[:;,\.\!\?\)\"\']$/
					|| ($def[$#def] !~ / $/ && $str =~ /^\(\"\'/)) {
				$str = " $str";
			} elsif ($def[$#def] =~ /[\x21-\x7e]$/
					&& $str =~ /^[\x21-\x7e]/) {
				$str = " $str";
			}
			$def[$#def] .= $str;
		} elsif ($str =~ /^/) {
			# ..., image block. this block ends by .
			push(@col, $str);
			while ($str = $file{HANDLE}->getline()) {
				last if ($str !~ s/^ {8}([^\r\n]*)[\r\n]+$/$1/);
				push(@col, $str);
				last if ($str =~ /^/);
			}
		} else {
			push(@col, $str);
		}
	}

	return(
		WORD => $word{WORD},
		READ => $info[0] ? $info[0] : '',
		ENGLISH => $info[1] ? $info[1] : '',
		PRONOUNCE => $info[2] ? $info[2] : '',
		POS => $info[3] ? $info[3] : '',
		DEFINITION => \@def,
		COLUMN => \@col,
		REFERENCE => \@ref,
		GROUP => $word{GROUP},
		FILE => $word{FILE},
		OFFSET => $word{OFFSET}
	);
}

#-----------------------------------------------------------------------#
#	index construction													#
#-----------------------------------------------------------------------#

sub Wdic::make_word_index()
{
	my $self = shift;
	my(@word_list, %word_index, %entry, $i, @key, $key);

	@word_list = @{ $self->{word_list} };

	for ($i = 0; $i < scalar(@word_list); $i++) {
		%entry = %{ $word_list[$i] };
		@key = unpack('SS', $entry{WORD});
		$key = sprintf("%04X", $key[0]);
		$key .= sprintf("%04X", $key[1]) if (defined($key[1]));
		push(@{ $word_index{$key} }, $i);
	}

	$self->{word_index} = \%word_index;
	return 1;
}

#-----------------------------------------------------------------------#
#	group file management												#
#-----------------------------------------------------------------------#

sub Wdic::load_group($)
{
	my $self = shift;
	my(%wdic_group, %wdic_file, %entry, $group, $handle, $section, $str);

	$handle = new IO::File($_[0], 'r')
		|| die "Unable to open the file '$_[0]' ($!)";
	&trace("Reading '$_[0]'...\n");
	binmode $handle;

	$group = uc($_[0]);
	$group =~ s/^.*\/([^\/]+)\.WLF$/$1/;

	%wdic_group = %{ $self->{wdic_group} };
	%wdic_file = %{ $self->{wdic_file} };

	while ($str = $handle->getline()) {
		$str =~ s/[\r\n]+$//;
		$str =~ s/\s*\#.*$//;
		next if ('' eq $str);

		if ($str =~ s/^\[(.+)\]$/$1/) {		# section declaration.
			$section = $str;

		} elsif ($section eq 'GROUP') {		# in [GROUP] section.
			next if ($str !~ s/^([^\=]+)\=//);
			if ('NAME' eq $1 || 'CONTENT' eq $1) {
				$entry{$1} = $str;
			} elsif ('CONTACT' eq $1) {
				($entry{CONTACT}, $entry{EMAIL}) =
					split(/\t/, $str);
			}

		} elsif ($section eq 'FILES') {		# dictionary entry.
			my(%file_entry, @info);

			@info = split(/\t/, $str);
			$info[0] =~ s/\.DIC$//i;
			%file_entry = (
				GROUP => $group,
				NAME => $info[1] ? $info[1] : '',
				CONTENT => $info[2] ? $info[2] : ''
			);
			push(@{ $entry{FILES} }, $info[0]);
			$wdic_file{$info[0]} = \%file_entry;
		}
	}

	$handle->close();
	%entry || die "The file '$_[0]' is not a WDIC group file";

	# update the class fields.
	$wdic_group{$group} = \%entry;
	$self->{wdic_group} = \%wdic_group;
	$self->{wdic_file} = \%wdic_file;
	return 1;
}

#-----------------------------------------------------------------------#
#	dictionary file management											#
#-----------------------------------------------------------------------#

sub Wdic::load_file($)
{
	my $self = shift;
	my(%wdic_file, %entry, $file, $handle, $capital, $str, $wc);

	$handle = new IO::File($_[0], 'r')
		|| die "Unable to open the file '$_[0]' ($!)";
	&trace("Indexing '$_[0]'...\n");
	binmode $handle;

	$file = uc($_[0]);
	$file =~ s/^.*\/([^\/]+)\.DIC$/$1/;

	%wdic_file = %{ $self->{wdic_file} };
	%entry = %{ $wdic_file{$file} } if ($wdic_file{$file});

	if (!%entry) {
		%entry = (
			GROUP => '',
			NAME => '',
			CONTENT => '',
			CONTACT => '',
			EMAIL => ''
		);
		warn "'$file' is not contained in any group";
	}
	if (!$entry{GROUP} || !$entry{NAME}) {
		# determine the dictionary name.
		while ($str = $handle->getline()) {
			last if ($str =~ /^[\<\-%]/);
			if ($str =~ /^\s*($SJIS_ANY+(?:̊bm)?)\s*(?:\x82[\x4f-\x58])+\s*(.+)[\r\n]+$/) {
				$entry{GROUP} = $1 if (!$entry{GROUP});
				$entry{NAME} = $2 if (!$entry{NAME});
				last;
			}
		}
	}

	# construct the word list.
	$capital = '';
	$wc = 0;

	while ($str = $handle->getline()) {
		if ($str =~ s/^%([^\r\n]+)[\r\n]+$/$1/) {		# %...
			if ($str eq 'L') {
				$capital = '0000';
			} elsif ($str eq '') {
				$capital = '0001';
			} else {
				$capital = sprintf("%02x%02x", unpack('CC', $str));
			}

		} elsif ($str =~ /^\#([^\r\n]+)[\r\n]+$/) {		# #...
			my %word = (
				WORD => $1,
				SORT => '',
				GROUP => $entry{GROUP},
				FILE => $file,
				OFFSET => $handle->tell() - length($str)
			);
			my($info, $read);

			if ($word{OFFSET} <= 0) {
				die "Read error on '$_[0]' ($!)";
			}

			# read information line(s).
			$str = $handle->getline();
			if ($str !~ s/^ {8}([^\r\n]+)[\r\n]+$/$1/) {
				warn "$word{WORD}: Information line not found";
				next;
			}

			$info = $str;
			while ($info !~ /l$/) {
				$str = $handle->getline();
				last if (!$str
					|| $str !~ s/^ {8}([^ ][^\r\n]*)[\r\n]+$/$1/
					|| $str =~ /^\x81[\x45\xa8]/);		# /^[E]/
				$info .= $str;
			}

			# store the information into a hash.
			$read = ($info !~ /^\[((?:$SJIS_ANY|[^\]])+)\]/) ? $1 : '';
			if ($capital =~ /^00/
					|| $capital =~ /^82[67]/
					|| $word{WORD} =~ /^$SJIS_YOMI+$/
					|| !$read) {
				$word{SORT} = $capital . make_sort_word($word{WORD});
			} else {
				$word{SORT} = $capital . make_sort_word($read);
			}

			# push this element to the list.
			push(@{ $self->{word_list} }, \%word);
			$wc++;
		}
	}

	if (!$wc) {
		$handle->close();
		die "No word found in the file '$_[0]'";
	}

	# update the class field.
	$entry{HANDLE} = $handle;
	$wdic_file{$file} = \%entry;
	$self->{wdic_file} = \%wdic_file;
	return $wc;
}

#-----------------------------------------------------------------------#
#	local functions														#
#-----------------------------------------------------------------------#
#
# char *make_sort_word(char *reading)
# converts Shift JIS characters, hiragana to katakana, lower to upper
# for case-insensitive sorting.
#
sub make_sort_word($)
{
	my(@chars) = unpack('C*', $_[0]);
	my($str, $c, $c2);

	$str = '';
	while ($c = shift(@chars)) {
		if (0x81 == $c) {
			$c2 = shift(@chars);
			if (0x40 == $c2 || 0x5b == $c2 || 0x7c == $c2) {
				# [@[|], nop.
			} else {
				$str .= pack('CC', $c, $c2);
			}
		} elsif (0x82 == $c) {
			$c2 = shift(@chars);
			if (0x81 <= $c2 && $c2 <= 0x9a) {
				$c2 -= 0x21;
			} elsif (0x9f <= $c2 && $c2 <= 0xdd) {
				$c = 0x83;
				$c2 -= 0x5f;
			} elsif (0xde <= $c2 && $c2 <= 0xf1) {
				$c = 0x83;
				$c2 -= 0x5e;
			}
			$str .= pack('CC', $c, $c2);
		} elsif ((0x83 <= $c && $c <= 0x9f) || (0xe0 <= $c && $c <= 0xfc)) {
			$c2 = shift(@chars);
			$str .= pack('CC', $c, $c2);
		} elsif (0x20 == $c || 0x2d == $c) {
			# [ -], nop.
		} else {
			$str .= pack('C', $c);
		}
	}
	return($str);
}

#
# char **split_info(char *info)
# extracts reading, english, pronounce and POS from the word information
# line.
#
sub split_info($)
{
	my(@chars, @info, $str, $paren, $c, $i);

	$str = $_[0];
	@info = ('', '', '', '');

	if ($str =~ /^\[((?:$SJIS_ANY|[^\]])+)\]/) {
		$info[0] = $1;
	}
	if ($str =~ s/k((?:$SJIS_ANY|[@\+\/])+)l//) {
		#
		#k...l, the part of speech. '/' is not a valid char here, but
		# contained in the RC5 version of APPL.DIC..
		#
		$info[3] = $1;
	}

	@chars = unpack('C*', $str);
	$str = '';
	$paren = 0;

	for ($i = 0; $i < scalar(@chars); $i++) {
		$c = $chars[$i];
		if ((0x81 <= $c && $c <= 0x9f) || (0xe0 <= $c && $c <= 0xfc)) {
			$i++;
			if ($paren) {
				$info[1] .= pack('CC', $c, $chars[$i]);
			} elsif ('/' eq $str) {
				$info[2].= pack('CC', $c, $chars[$i]);
			}
			next;
		}
		if ('/' eq $str) {
			last if (0x2f == $c);		# /
			$info[2] .= pack('C', $c);
		} elsif (!$paren && '' eq $str && 0x2f == $c) {
			$str = '/';
		} else {
			if (0x28 == $c) {			# (
				$paren++;
				next if (1 == $paren);
			} elsif (0x29 == $c) {		# )
				$paren--;
				next if (!$paren);
			}
			$info[1] .= pack('C', $c) if (0 < $paren);
		}
	}

	$info[2] =~ s/ *[,\|] +\- +/\|-/g;
	$info[2] =~ s/ *[,\|] +/\|/g;
	$info[2] =~ s/ +(\[$SJIS_ANY\])/$1/g;

	return(@info);
}

#-----------------------------------------------------------------------#
#	debugging functions													#
#-----------------------------------------------------------------------#

sub trace($)
{
	my $str = shift;

	return if (!$TRACE);
	$str = sprintf($str, @_) if (@_);
	if ($str =~ /[^\x00-\x7e]/ && '' ne $CHARSET && $INC{'Jcode.pm'}) {
		$str = Jcode::convert($str, $CHARSET);
	}
	print $str;
}

sub dump_array($$@)
{
	my($indent, $name, @array) = @_;
	my($i, $line);

	print ' ' x $indent . "\@$name =";
	if (!@array) {
		print " (empty)\n";
		return;
	}
	print "\n";

	for ($i = 0; $i < scalar(@array); $i++) {
		if (!defined($array[$i])) {
			$line = '(undef)';
		} elsif (ref($array[$i]) =~ "HASH") {
			&dump_hash($indent + 2, "[$i]", %{ $array[$i] });
			next;
		} elsif (ref($array[$i]) eq "ARRAY") {
			&dump_array($indent + 2, "[$i]", @{ $array[$i] });
			next;
		} else {
			$line = $array[$i];
		}
		if ($line =~ /[^\x00-\x7e]/
				&& '' ne $CHARSET
				&& defined($INC{'Jcode.pm'})) {
			$line = Jcode::convert($line, $CHARSET);
		}
		print ' ' x ($indent + 2) . "[$i] = $line\n";
	}
}

sub dump_hash($$%)
{
	my($indent, $name, %hash) = @_;
	my($key, $value, $line);

	print ' ' x $indent . "\%$name =";
	if (!%hash) {
		print " (empty)\n";
		return;
	}
	print "\n";

	while (($key, $value) = each(%hash)) {
		if (!defined($value)) {
			$line = '(undef)';
		} elsif (ref($value) =~ "HASH") {
			&dump_hash($indent + 2, $key, %{ $value });
			next;
		} elsif (ref($value) eq "ARRAY") {
			&dump_array($indent + 2, $key, @{ $value });
			next;
		} else {
			$line = $value;
		}
		if ($line =~ /[^\x00-\x7e]/
				&& '' ne $CHARSET
				&& defined($INC{'Jcode.pm'})) {
			$line = Jcode::convert($line, $CHARSET);
		}
		print ' ' x ($indent + 2) . "$key => $line\n";
	}
}

1;
